var Fd = {
  IMAGE_BLANK: FdFile.getAsDataURL("chrome://fastdial/skin/blank.gif"),
  menu: null,
  addMenu: null,
  WIDTH: FdPrefs.getIntPref("width"),
  HEIGHT: FdPrefs.getIntPref("height"),
  currentId: 0,
  currentTarget: null,
  pageInfos: new Array(), // cached pageInfos
  prefsWindow: null,
  RENDER_TIMEOUT: 500,
  forceDrawPage: false,
  pageHtml: null,

  initialize: function() {
    Fd.initShortcutKeys();

    Fd.menu = $("fastdial-menu");
    Fd.addMenu = $("fastdial-add-menu");
    $("contentAreaContextMenu").addEventListener("popupshowing",
                                                 Fd.onContextShowing, false);
    gBrowser.addEventListener("load", Fd.onPageLoad, true);

    for(var i = 1; i <= Fd.WIDTH * Fd.HEIGHT; i++) {
      Fd.pageInfos[i] = FdPrefs.getPageInfo(i);
      if (Fd.pageInfos[i].refresh > 0) Fd.refreshThumbnail(i);
    }

    // Tabs' "busy" attribute may hasn't been set yet, let some time to pass
    setTimeout(Fd.drawPagesOnStartup, 0);
  },

  drawPagesOnStartup: function() {
    FdUtils.forEachBlankPage(function(doc) {
      var tab = FdUtils.getDocumentTab(doc);
      if (!tab.hasAttribute("busy") && FdUtils.isBlank(doc)) {
        Fd.drawPage(doc);
      }
    });
  },

  initShortcutKeys: function() {
    var modifiers = FdPrefs.getCharPref("shortcutKeys");
    for(var i = 1; i <= 9; i++) {
      var key = $("fastdial-key-" + i);
      key.setAttribute("modifiers", modifiers);
    }
  },

  open: function(event) {
    if (event.button == 0) {
      Fd.forceDrawPage = true;
      FdUtils.openUILink("about:blank");
    }
    else if (event.button == 1) {
      Fd.forceDrawPage = true;
      FdUtils.openUILink("about:blank", "tab");
    }
  },

  onPageLoad: function(event) {
    var doc = event.originalTarget;
    if (!FdUtils.getDocumentTab(doc)) return; // not top-level document

    if (FdUtils.isBlank(doc) &&
        (Fd.forceDrawPage ||
         gBrowser.browsers.length == 1 ||
         FdPrefs.getBoolPref("showInNewTabs")))
    {
      Fd.forceDrawPage = false;
      Fd.drawPage(doc);
    }
    else {
      // Refresh thumbnail for this page if there's refreshOnVisit specified
      for(var i = 1; i <= Fd.WIDTH * Fd.HEIGHT; i++) {
        var pageInfo = Fd.pageInfos[i];
        if (pageInfo.refreshOnVisit &&
            doc.location.href == pageInfo.url)
        {
          Fd.saveDoc(i, doc);
        }
      }
    }
  },

  drawPage: function(doc) {
    if (!Fd.pageHtml) {
      Fd.pageHtml = Fd.createPageHtml();
    }
    doc.body.innerHTML = Fd.pageHtml;
    Fd.updateDoc(doc);

    doc.addEventListener("click", FdPage.onClick, false);
    doc.addEventListener("mousemove", FdPage.onMouseMove, false);
    doc.addEventListener("mousedown", FdPage.onMouseDown, false);
    doc.addEventListener("mouseup", FdPage.onMouseUp, false);

    FdURL.addToHistory("about:blank");
  },

  invalidatePageHtml: function() {
    Fd.pageHtml = null;
  },

  createPageHtml: function() {
    Fd.WIDTH = FdPrefs.getIntPref("width");
    Fd.HEIGHT = FdPrefs.getIntPref("height");

    var hide = " style=\"display: none;\"";

    var html = "";
    for(var i = 1; i <= Fd.WIDTH * Fd.HEIGHT; i++) {
      if (!Fd.pageInfos[i]) Fd.pageInfos[i] = FdPrefs.getPageInfo(i);
      var pageInfo = Fd.pageInfos[i];
      var title = pageInfo.title || "&nbsp;";

      if (i % Fd.WIDTH == 1) html += "<tr>";
      html += "<td><div class=\"cell\" id=\"cell" + i + "\">" +
              "<div class=\"buttons\" id=\"buttons" + i + "\">" +
              "<div class=\"button clear\" title=\"" +
              FdBundle.getString("clear") + "\"></div>" +
              "<div class=\"button refresh\" title=\"" +
              FdBundle.getString("refresh") + "\"></div>" +
              "<div class=\"button setURL\" title=\"" +
              FdBundle.getString("setURL") + "\"></div>" +
              "</div>" +
              "<div class=\"loading\" id=\"loading" + i + "\"" +
              (pageInfo._loading ? "" : hide) + "></div>" +
              "<a href=\"" + pageInfo.url + "\" id=\"link" + i + "\"" +
              (pageInfo._loading ? hide : "") + ">" +
              "<img class=\"image\" id=\"" + i + "\" src=\"" + Fd.getImageURL(i) + "\"></a>" + 
              "<div class=\"title\" id=\"title" + i + "\">" + title + "</div>" +
              "</div></td>";
      if (i % Fd.WIDTH == 0) html += "</tr>";
    }
    return FdTheme.current.template.replace(/%content%/, html);
  },

  openLink: function(event) {
    var id = FdPanel.getId(event.target);
    var pageInfo = Fd.pageInfos[id];
    FdUtils.openUILink(pageInfo.url, pageInfo.openIn || event);
  },

  exchangePanels: function(sourceId, targetId) {
    // Exchange pageInfos
    var targetPageInfo = Fd.pageInfos[sourceId];
    var sourcePageInfo = Fd.pageInfos[targetId];
    targetPageInfo.id = targetId;
    sourcePageInfo.id = sourceId;
    Fd.pageInfos[sourceId] = sourcePageInfo;
    Fd.pageInfos[targetId] = targetPageInfo;
    FdPrefs.savePageInfo(sourcePageInfo);
    FdPrefs.savePageInfo(targetPageInfo)

    // Exchange files
    FdFile.exchangeImages(sourceId, targetId);

    // Update images
    Fd.updateImage(sourceId);
    Fd.updateImage(targetId);
  },

  onContextShowing: function(event) {
    if (event.target.id == "contentAreaContextMenu") {
      var isBlankDoc = FdUtils.isBlank(content.document);
      Fd.menu.hidden = !isBlankDoc ||
                       !gContextMenu.target ||
                       !(Fd.currentId = FdPanel.getId(gContextMenu.target));

      Fd.addMenu.hidden = isBlankDoc ||
                          !FdPrefs.getBoolPref("showAddMenu");

      Fd.currentTarget = gContextMenu.target.href ||
                         gContextMenu.target.src;
    }
    else if (event.target.id == "fastdial-popup") {
      var url = Fd.pageInfos[Fd.currentId].url;
      $("fastdial-popup-refresh").hidden = !url;
      $("fastdial-popup-clear").hidden = !url;
    }
    else if (event.target.id == "fastdial-add-popup") {
      FdUtils.clearElement(event.target);
      for(var i = 1; i <= Fd.WIDTH * Fd.HEIGHT; i++) {
        var menuitem = document.createElement("menuitem");
        menuitem.setAttribute("label", i + "." + Fd.pageInfos[i].title);
        menuitem.setAttribute("class", "menuitem-iconic");
        menuitem.setAttribute("image", FdURL.getFaviconURL(Fd.pageInfos[i].url));
        menuitem.setAttribute("oncommand", "Fd.addTarget(" + i + ");");
        event.target.appendChild(menuitem);
      }
    }
  },

  pageInfo: null,

  setURL: function(id) {
    var oldPageInfo = Fd.pageInfos[id];
    if (!Fd.pageInfo) Fd.pageInfo = FdUtils.clone(Fd.pageInfos[id]);

    openDialog("chrome://fastdial/content/seturl.xul",
               "fastdial-seturl-window",
               "chrome,centerscreen=yes,modal=yes,resizable=yes");

    if (Fd.pageInfo) {
      if (Fd.pageInfo.url) {

        Fd.pageInfos[id] = Fd.pageInfo;
        FdPrefs.savePageInfo(Fd.pageInfo);

        var doc = content.document;
        if (!Fd.pageInfo.customImage &&
           Fd.pageInfo.url == doc.location.href)
        {
          Fd.saveDoc(id, doc);
        }
        else if (Fd.pageInfo.url != oldPageInfo.url || !Fd.pageInfo.title ||
                 Fd.pageInfo.customImage != oldPageInfo.customImage ||
                 Fd.pageInfo.customBackground != oldPageInfo.customBackground)
        {
          Fd.refreshThumbnail(id);
        }
        else {
          if (Fd.pageInfo.title != oldPageInfo.title) {
            Fd.updateImage(id);
          }
          if (Fd.pageInfo.refresh != oldPageInfo.refresh) {
            Fd.updateRefreshTimeout(id);
          }
        }
      } else Fd.clearURL(id, true);
    }
    Fd.pageInfo = null;
  },

  refreshThumbnail: function(id) {
    var pageInfo = Fd.pageInfos[id];
    if (!pageInfo.url) return;

    pageInfo._loading = true;
    Fd.updateImage(id);

    if (pageInfo.customImage && pageInfo.title) {
      Fd.saveCustomImage(pageInfo);
    }
    else Fd.saveImage(pageInfo);
  },

  createFrame: function(url) {
    var doc = content.document;
    // Local URLs cannot be accessed from "about:blank" page, so use hidden window
    if (FdURL.isLocal(url) || !FdUtils.isBlank(doc)) {
      doc = FdUtils.getHiddenWindow().document;
    }
    var iframe = doc.createElement("iframe");
    iframe.style.position = "absolute";
    iframe.style.left = -2000; // Position the frame outside the screen
    iframe.style.top = -2000;
    iframe.width = 1024;
    iframe.height = 768;
    doc.body.appendChild(iframe);
    return iframe;
  },

  deleteFrame: function(iframe) {
    iframe.parentNode.removeChild(iframe);
  },

  loadInFrame: function(url, onLoad) {
    function onFrameLoad() {
      clearTimeout(loadTimeout);
      onLoad(iframe);
    }
    var iframe = Fd.createFrame(url);
    iframe.addEventListener("load", onFrameLoad, true);
    var loadTimeout = setTimeout(onFrameLoad,
                                 FdPrefs.getIntPref("maxLoadTime") * 1000);
    try {
      iframe.setAttribute("src", url);
    }
    catch(e) {
      FdUtils.error("incorrectURL", [url]);
      onFrameLoad();
    }
  },

  saveImage: function(pageInfo) {
    Fd.loadInFrame(pageInfo.url, function(iframe) {
        if (!pageInfo.title) {
          var doc = iframe.contentDocument;
          pageInfo.title = doc.title || doc.location.href;
          FdPrefs.savePageInfo(pageInfo);
        }
        if (pageInfo.customImage) {
          Fd.deleteFrame(iframe);
          Fd.saveCustomImage(pageInfo);
        }
        else Fd.saveThumbnail(pageInfo.id, iframe);
      });
  },

  saveCustomImage: function(pageInfo) {
    var url = pageInfo.customImage;
    FdURL.removeFromCache(url);

    Fd.loadInFrame(url, function(iframe) {
        var doc = iframe.contentDocument;
        var img = doc.body.firstChild;
        if (FdThumbnail.getHeight(img.width) / img.height > 1) {
          iframe.width = img.width;
          iframe.height = FdThumbnail.getHeight(img.width);
        }
        else {
          iframe.width = FdThumbnail.getWidth(img.height);
          iframe.height = img.height;
        }
        doc.body.style.background = pageInfo.customBackground || "white";
        doc.body.style.margin = 0;
        doc.body.style.width = iframe.width;
        doc.body.style.height = iframe.height;
        doc.body.style.display = "table-cell";
        doc.body.style.textAlign = "center";
        doc.body.style.verticalAlign = "middle";

        Fd.saveThumbnail(pageInfo.id, iframe);
      });
  },

  saveThumbnail: function(id, iframe) {
    Fd.pageInfos[id]._loading = false;
    Fd.pageInfos[id]._imageURL = null;

    setTimeout(function() {
        var thumbWidth = FdPrefs.getIntPref("thumbWidth");
        FdThumbnail.createThumbnail(iframe.contentWindow, thumbWidth,
                                    function(image) {
                                      Fd.deleteFrame(iframe);
                                      FdFile.saveImage(id, image);
                                      Fd.updateImage(id);
                                      Fd.updateRefreshTimeout(id);
                                    });
      },
      Fd.RENDER_TIMEOUT);
  },

  saveDoc: function(id, doc) {
    var iframe = Fd.createFrame();
    iframe.addEventListener("load", function() {
        Fd.saveThumbnail(id, iframe);
      },
      true);

    var nsiUrl = FdURL.getNsiURL(doc.location.href);
    nsiUrl.fileName = "";

    iframe.src = "data:text/html,<base href=\"" + nsiUrl.spec + "\">" +
                 "<META http-equiv=\"Content-Type\" content=\"text/html; charset=utf-8\">" +
                 doc.documentElement.innerHTML;
  },

  updateImage: function(id) {
    FdUtils.forEachBlankPage(function(doc) {
      var image = doc.getElementById(id);
      if (image) {
        var pageInfo = Fd.pageInfos[id];

        var url = Fd.getImageURL(id)
        FdURL.removeFromCache(url);
        image.src = url;

        var panel = new FdPanel(image);
        var link = panel.get("link");
        link.href = pageInfo.url;
        link.style.display = pageInfo._loading ? "none" : "block";

        panel.get("loading").style.display = pageInfo._loading ? "block" : "none";
        panel.get("title").innerHTML = pageInfo.title || "&nbsp;";
      }
    });
    Fd.invalidatePageHtml();
  },

  clearURL: function(id, dontAskConfirmation) {
    var pageInfo = Fd.pageInfos[id];
    if (!pageInfo.url) return;

    var title = pageInfo.title || pageInfo.url;
    var message = FdBundle.getString("clearConfirmation", [title]);
    if (dontAskConfirmation || FdUtils.confirm(message)) {
      clearTimeout(pageInfo._timeout);
      FdPrefs.clearPageInfo(id);
      Fd.pageInfos[id] = FdPrefs.getPageInfo(id); // update pageInfo cache
      FdFile.deleteImage(id);
      Fd.updateImage(id);
    }
  },

  getImageSize: function() {
    var width, height;
    var style = FdTheme.current.style;

    if (FdPrefs.getBoolPref("fixedSize")) {
      width = FdPrefs.getIntPref("thumbWidth");
      height = FdThumbnail.getHeight(width);
    } else {
      var wnd = gBrowser.selectedBrowser.contentWindow;
      width = (wnd.innerWidth - (style.body.margin*2 + style.content.borderSpacing)) /
              Fd.WIDTH - (parseInt(style.cell.border)*2 + style.cell.padding*2 +
              parseInt(style.image.border)*2 + style.content.borderSpacing);

      height = (wnd.innerHeight - (style.body.margin*2 + style.content.borderSpacing)) /
               Fd.HEIGHT - (parseInt(style.cell.border)*2 + style.cell.padding +
               parseInt(style.image.border)*2 + style.content.borderSpacing +
               style.title.margin*2 + style.body.fontSize + 4);
    }
    width = Math.max(Math.floor(width), style.image.minWidth);
    height = Math.max(Math.floor(height), style.image.minHeight);

    if (FdThumbnail.getHeight(width) / height > 1) {
      width = FdThumbnail.getWidth(height);
    }
    else height = FdThumbnail.getHeight(width);

    return { width: width, height: height };
  },

  updatePage: function() {
    FdUtils.forEachBlankPage(Fd.updateDoc);
  },

  redrawPage: function() {
    FdUtils.forEachBlankPage(Fd.drawPage);
  },

  updateDoc: function(doc) {
    var imageSize = Fd.getImageSize();
    var style = FdTheme.current.style;

    for(var i = 0; i < doc.styleSheets.length; i++) {
      var styleSheet = doc.styleSheets[i];
      if (styleSheet.href &&
          styleSheet.href != "about:blank") continue;

      var rules = styleSheet.cssRules;
      for(var j = 0; j < rules.length; j++) {
        var rule = rules[j];
        switch(rule.selectorText) {
          case ".loading":
          case ".image":
              rule.style.height = imageSize.height;
              rule.style.width = imageSize.width;
              break;
          case ".title":
              rule.style.width = imageSize.width +
                                 parseInt(style.image.border)*2 - style.title.margin*2;
              break;
        }
        Fd.applyStyle(rule, style);
      }
    }
  },

  applyStyle: function(rule, style) {
    if (!rule.selectorText) return;

    var regexp = /\s*([^#.]+)\s*(,|$)/g;
    while(match = regexp.exec(rule.selectorText)) {
      var entry = style[match[1]];
      if (!entry) continue;

      if (rule.selectorText.match(/hover/)) {
        entry = entry.hover;
      }
      for(var i in entry) {
        rule.style[i] = entry[i];
      }
    }
  },

  openPreferences: function() {
    if (!Fd.prefsWindow || Fd.prefsWindow.closed) {
      Fd.prefsWindow = window.openDialog("chrome://fastdial/content/preferences.xul",
                                         "fastdial-preferences-window",
                                         "chrome,toolbar,centerscreen,resizable=yes");
    } else Fd.prefsWindow.focus();
  },

  getImageURL: function(id) {
    var pageInfo = Fd.pageInfos[id];

    if (pageInfo.url && pageInfo.imageURL == null) {
      pageInfo._imageURL = FdFile.getImageURL(id);
    }
    var url = pageInfo._imageURL || Fd.IMAGE_BLANK;
    return url;
  },

  openPage: function(id) {
    var pageInfo = FdPrefs.getPageInfo(id);
    FdUtils.openUILink(pageInfo.url);
  },

  updateRefreshTimeout: function(id) {
    var pageInfo = Fd.pageInfos[id];
    if (pageInfo._timeout) clearTimeout(pageInfo._timeout);
    if (pageInfo.refresh > 0) {
      pageInfo._timeout = setTimeout(function() { Fd.refreshThumbnail(pageInfo.id); },
                                     pageInfo.refresh * 60 * 1000);
    }
  },

  addTarget: function(id) {
    var doc = content.document;
    var customImage = FdURL.isImage(Fd.currentTarget) && Fd.currentTarget;
    if (customImage) {
      Fd.pageInfo = FdUtils.clone(Fd.pageInfos[id]);
      Fd.pageInfo.customImage = customImage;
    } else {
      Fd.pageInfo = new FdPageInfo(id);
      Fd.pageInfo.url = Fd.currentTarget || doc.location.href;
      if (!Fd.currentTarget) Fd.pageInfo.title = doc.title;
    }
    Fd.setURL(id);
  }
}
