var FdFile = {
  getProfileDirectory: function() {
    var file = Components.classes["@mozilla.org/file/directory_service;1"]
               .getService(Components.interfaces.nsIProperties)
               .get("ProfD", Components.interfaces.nsIFile);
    file.append("fastdial");
    if (!file.exists()) {
      file.create(Components.interfaces.nsIFile.DIRECTORY_TYPE, 0777);
    }
    return file;
  },

  getExtensionDirectory: function() {
    var id = "fastdial@telega.phpnet.us";
    return Components.classes["@mozilla.org/extensions/manager;1"]
           .getService(Components.interfaces.nsIExtensionManager)
           .getInstallLocation(id)
           .getItemLocation(id);
  },

  getPredefinedThemesDirectory: function() {
    var file = FdFile.getExtensionDirectory();
    file.append("chrome");
    file.append("content");
    file.append("themes");
    return file;
  },

  getCustomThemesDirectory: function() {
    var file = FdFile.getProfileDirectory();
    file.append("themes");
    if (!file.exists()) {
      file.create(Components.interfaces.nsIFile.DIRECTORY_TYPE, 0777);
    }
    return file;
  },

  getImageName: function(id) {
    return id + ".png";
  }, 

  getImageFile: function(id) {
    var file = FdFile.getProfileDirectory();
    file.append(FdFile.getImageName(id));
    return file;
  },

  deleteImage: function(id) {
    var file = FdFile.getProfileDirectory();
    file.append(FdFile.getImageName(id));
    try {
      file.remove(false);
    } catch(e) {}
  },

  // "chrome://" images in FF 2.x expire from cache after 1 minute,
  // so use "data:" urls to avoid reloading of images

  getImageURL: function(id) {
    var url = "chrome://fastdial-profile/content/" + FdFile.getImageName(id);
    try {
      return FdUtils.isFirefox3() ? url : FdFile.getAsDataURL(url);
    }
    catch(e) { return ""; }
  },

  saveImage: function(id, data) {
    var file = FdFile.getImageFile(id);
    FdFile.writeFile(file, data);
  },

  writeFile: function(file, data) {
    var out = Components.classes["@mozilla.org/network/file-output-stream;1"]
              .createInstance(Components.interfaces.nsIFileOutputStream);
    out.init(file, 0x04 | 0x08 | 0x20, 0666, 0); // read & write, create, truncate
    out.write(data, data.length);
    out.close();
  },

  exchangeImages: function(sourceId, targetId) {
    var sourceFile = FdFile.getImageFile(sourceId);
    var sourceExists = sourceFile.exists();
    var targetFile = FdFile.getImageFile(targetId);
    var targetExists = targetFile.exists();
    if (sourceExists && targetExists) {
      sourceFile.moveTo(null, "tmp" + new Date().getTime());
    }
    if (targetExists) targetFile.moveTo(null, FdFile.getImageName(sourceId));
    if (sourceExists) sourceFile.moveTo(null, FdFile.getImageName(targetId));
  },

  getFileURL: function(file) {
    var nsiFile;
    if (file instanceof Components.interfaces.nsIFile) nsiFile = file;
    else {
      nsiFile = Components.classes["@mozilla.org/file/local;1"]
                .createInstance(Components.interfaces.nsILocalFile);
      nsiFile.initWithPath(file);
    }
    var ios = Components.classes["@mozilla.org/network/io-service;1"]
              .getService(Components.interfaces.nsIIOService);
    var fileHandler = ios.getProtocolHandler("file")
                     .QueryInterface(Components.interfaces.nsIFileProtocolHandler);
    return fileHandler.getURLSpecFromFile(nsiFile);
  },

  read: function(urlOrFile, unichar) {
    var url = (urlOrFile instanceof Components.interfaces.nsIFile)
                ? FdFile.getFileURL(urlOrFile)
                : urlOrFile;

    var ioService = Components.classes["@mozilla.org/network/io-service;1"]
                    .getService(Components.interfaces.nsIIOService);
    var channel = ioService.newChannel(url, null, null);
    var stream = channel.open();

    function readUnicharStream(stream) {
      var unichar = Components.classes["@mozilla.org/intl/converter-input-stream;1"]
                    .createInstance(Components.interfaces.nsIConverterInputStream);
      unichar.init(stream, "utf-8", stream.available(), 0xFFFD);
      var str = {};
      unichar.readString(stream.available(), str);
      unichar.close();
      return str.value;
    }

    function readBinaryStream(stream) {
      var binary = Components.classes["@mozilla.org/binaryinputstream;1"]
                   .createInstance(Components.interfaces.nsIBinaryInputStream);
      binary.setInputStream(stream);
      var bytes = binary.readBytes(binary.available());
      binary.close();
      return bytes;
    }

    var data = unichar ? readUnicharStream(stream)
                       : readBinaryStream(stream);
    stream.close();
    return data;
  },

  getAsDataURL: function(urlOrFile) {
    var data = FdFile.read(urlOrFile);
    return "data:image/png;base64," + btoa(data);
  }
}

var FdURL = {
  getNsiURL: function(url) {
    var nsiUrl = Components.classes["@mozilla.org/network/standard-url;1"]
                 .createInstance(Components.interfaces.nsIURL);
    nsiUrl.spec = url;
    return nsiUrl;
  },

  getFaviconURL: function(url) {
    if (url) {
      var nsiUrl = FdURL.getNsiURL(url);
      if (nsiUrl.scheme != "ftp") {
        return nsiUrl.prePath + "/favicon.ico";
      }
    }
  },

  getScheme: function(url) {
    if (url) {
      return FdURL.getNsiURL(url).scheme;
    }
  },

  isLocal: function(url) {
    return url && url.match(/^(file:|chrome:)/);
  },

  isImage: function(url) {
    return url && url.match(/\.(jpg|jpeg|png|gif|bmp)$/);
  },

  removeFromCache: function(url) {
    if (!url) return;

    var clientId = url.match(/^chrome:/) ? "image-chrome" : "image";
    var cacheSession = Components.classes["@mozilla.org/network/cache-service;1"]
                       .getService(Components.interfaces.nsICacheService)
                       .createSession(clientId, 0, false);
    try {
      var entry = cacheSession.openCacheEntry(url,
                    Components.interfaces.nsICache.ACCESS_READ, false);
      entry.doom();
    }
    catch(e) {}
  },

  addToHistory: function(url) {
    var sessionHistory = gBrowser.sessionHistory;
    sessionHistory.QueryInterface(Components.interfaces.nsISHistoryInternal);

    if (!sessionHistory.count) {
      // nsIURL doesn't work for "about:blank" url
      var nsiUri = Components.classes["@mozilla.org/network/simple-uri;1"]
                   .createInstance(Components.interfaces.nsIURI);
      nsiUri.spec = url;
      var entry = Components.classes["@mozilla.org/browser/session-history-entry;1"]
                  .createInstance(Components.interfaces.nsISHEntry);
      entry.setURI(nsiUri);
      sessionHistory.addEntry(entry, true);
    }
  }
}
