function FdThemePref(id, url) {
  this.id = id;
  this.url = url || FdTheme.PREDEFINED_URL + this.id;
  this.style = { body: {},
                 content: {},
                 cell: { hover: {} },
                 image: { hover: {} }
               };
}

FdTheme = { 
  PREDEFINED_URL: "chrome://fastdial/content/themes/",
  CUSTOM_URL: "chrome://fastdial-profile/content/themes/",
  _themes: null,
  _current: null,

  get themes() {
    if (!FdTheme._themes) {
      FdTheme._themes = FdUtils.merge(FdTheme.loadPredefinedThemes(),
                                      FdTheme.loadCustomThemes());
    }
    return FdTheme._themes;
  },

  get current() {
    if (!FdTheme._current) {
      var themePref = FdTheme.getPref();
      var theme = FdTheme.loadTheme(themePref.id, themePref.url);
      FdTheme._current = FdUtils.merge(theme, themePref);
    }
    return FdTheme._current;
  },

  preload: function() {
    FdTheme.current;
  },

  getPref: function() {
    var themePref = FdPrefs.getObjectPref("theme");
    if (!themePref || !themePref.style) {
      themePref = new FdThemePref("default");
    }
    return themePref;
  },

  savePref: function(themePref) {
    themePref.style.buttons = themePref.style.cell;
    themePref.style.loading = themePref.style.image;

    FdPrefs.setObjectPref("theme", themePref);
    FdTheme._current = FdUtils.merge(FdTheme.themes[themePref.id], themePref);
  },

  loadTheme: function(id, url) {
    try {
      var template = FdFile.read(url + "/template.html", true);
      var style = eval(FdFile.read(url + "/style.js", true));

      return { id: id,
               url: url,
               template: template,
               style: style,
               name: FdBundle.getString(id + "Theme") || style.name || id
             };
    }
    catch(e) {
      FdUtils.error("themeError", [id, e]);
    }
  },

  loadPredefinedThemes: function() {
    return FdTheme.loadThemes(FdFile.getPredefinedThemesDirectory(),
                              FdTheme.PREDEFINED_URL);
  },

  loadCustomThemes: function() {
    return FdTheme.loadThemes(FdFile.getCustomThemesDirectory(),
                              FdTheme.CUSTOM_URL);
  },

  loadThemes: function(baseDir, baseUrl) {
    var themes = new Array();
    var dirs = baseDir.directoryEntries;

    while(dirs.hasMoreElements()) {
      var dir = dirs.getNext();
      dir.QueryInterface(Components.interfaces.nsIFile);

      if (dir.isDirectory()) {
        var id = dir.leafName;
        var theme = FdTheme.loadTheme(id, baseUrl + id);
        if (theme) themes[id] = theme;
      }
    }
    return themes;
  }
}
