function $(id) {
  return document.getElementById(id);
}

var FdUtils = {
  clearElement: function(element) {
    for(var i = element.childNodes.length - 1; i >= 0; i--) {
      element.removeChild(element.childNodes[i]);
    }
  },

  clone: function(object) {
    return FdUtils.copy(new Object(), object);
  },

  merge: function(object1, object2) {
    var result = FdUtils.clone(object1);
    return FdUtils.copy(result, object2);
  },

  // Deep copy. Source is merged into target.
  copy: function(target, source) {
    for(var i in source) {
      if (source[i] == null) continue;
      switch(source[i].constructor.name) {
        case "String":
        case "Number":
        case "Boolean":
            target[i] = source[i];
            break;
        default:
            target[i] = FdUtils.merge(target[i] || new Object(), source[i]);
            break;
      }
    }
    return target;
  },

  confirm: function(message) {
    var prompts = Components.classes["@mozilla.org/embedcomp/prompt-service;1"]
                  .getService(Components.interfaces.nsIPromptService);
    return prompts.confirm(window, FdBundle.getString("fastDial"), message);
  },

  getHiddenWindow: function() {
    return Components.classes["@mozilla.org/appshell/appShellService;1"]
           .getService(Components.interfaces.nsIAppShellService)
           .hiddenDOMWindow;
  },

  getBrowserWindow: function() {
    var wm = Components.classes["@mozilla.org/appshell/window-mediator;1"]
             .getService(Components.interfaces.nsIWindowMediator);
    return wm.getMostRecentWindow("navigator:browser");
  },

  getFontList: function() {
    var enumerator = Components.classes["@mozilla.org/gfx/fontenumerator;1"]
                     .getService(Components.interfaces.nsIFontEnumerator);
    var count = {};
    return enumerator.EnumerateAllFonts(count);
  },

  isBlank: function(doc) {
    return doc && doc.location && doc.location.href == "about:blank";
  },

  forEachBlankPage: function(onPage) {
    for (var i = 0; i < gBrowser.browsers.length; i++) {
      var doc = gBrowser.browsers[i].contentDocument;
      if (FdUtils.isBlank(doc)) onPage(doc);
    }
  },

  isFirefox3: function() {
    return Components.classes["@mozilla.org/login-manager;1"];
  },

  getDocumentTab: function(doc) {
    var tabs = gBrowser.tabContainer.childNodes;
    for(var i = 0; i < tabs.length; i++) {
      if (tabs[i].linkedBrowser.contentDocument == doc) {
        return tabs[i];
      }
    }
  },

  openUILink: function(url, where) {
    if (where instanceof Event &&
        !where.shiftKey &&
        !where.ctrlKey)
    {
      where = (where.button == 1) ? "tabshifted" : "";
    }

    // Firefox
    if (window.openUILink) {
      if (!where || where instanceof Event) {
        return openUILink(url, where)
      }
      else return openUILinkIn(url, where);
    }

    // Else SeaMonkey
    if (where instanceof Event) {
      return openNewTabOrWindow(where, url);
    }
    switch(where) {
      case "tabshifted": return gBrowser.addTab(url);
      case "tab":        return gBrowser.selectedTab = gBrowser.addTab(url);
      case "window":     return openNewWindowWith(url);
      default:           return loadURI(url);
    }
  },

  error: function(message, params) {
    try {
      message = FdBundle.getString(message, params);
    }
    catch(e) {}

    var fastDialError = FdBundle.getString("fastDialError");
    Components.utils.reportError(fastDialError + ": " + message);

    var prompts = Components.classes["@mozilla.org/embedcomp/prompt-service;1"]
                  .getService(Components.interfaces.nsIPromptService);
    return prompts.alert(window, fastDialError, message);
  },

  isJavaAvailable: function(onFailure) {
    var failureTimeout;
    if (onFailure) {
      failureTimeout = setTimeout(onFailure, 0);
    }
    // If Java isn't installed, the following command causes silent javascript
    // interruption without throwing any exception
    var result = window.Packages &&
                 window.Packages.java != null;
    if (onFailure) {
      clearTimeout(failureTimeout);
      if (!result) onFailure();
    }
    return result;
  },

  getLocale: function() {
    var localeService = Components.classes["@mozilla.org/intl/nslocaleservice;1"]
                        .getService(Components.interfaces.nsILocaleService);
    var locale = localeService.getApplicationLocale();
    return locale.getCategory("NSILOCALE_MESSAGES");
  },

  ajax: function(url, method, content, onLoad) {
    var request = new XMLHttpRequest();
    request.onreadystatechange = function() {
      if (request.readyState == 4) {
        if (request.status == 200) {
          onLoad(request.responseText);
        } else {
          FdUtils.error("serverError", [request.status + " " + request.statusText]);
          onLoad("");
        }
      }
    };
    request.open(method || "get", url, true);
    if (method == "post") {
      request.setRequestHeader("Content-Type", "application/x-www-form-urlencoded");
    }
    request.send(content);
  }
}

var FdBundle = {
  bundle: Components.classes["@mozilla.org/intl/stringbundle;1"]
          .getService(Components.interfaces.nsIStringBundleService)
          .createBundle("chrome://fastdial/locale/fastdial.properties"),

  getString: function(name, params) {
    return params ? FdBundle.bundle.formatStringFromName(name, params, params.length) 
                  : FdBundle.bundle.GetStringFromName(name);
  }
}
